#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright (C) 2009  B. Clausius
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Ported from GNUbik
# Original filename: move-queue.c
# Original copyright and license:
#/*
#  Copyright (c) 2004  Dale Mellor
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*/


#  OBJECT MoveQueue
#
#  DESCRIPTION
#
#    This object is designed to hold runs of moves which are applied to a
#    cube. New moves are pushed onto the back of the queue,  and old ones may be
#    popped of the front. The queue is indexed by a moving cursor (known
#    conventionally as the 'current' place) which can be moved backwards and
#    forwards. In this way,  the GNUbik application can maintain a running list
#    of moves being applied (either by the user or from scripts),  and can
#    allow the user to move backwards and forwards through the moves giving a
#    rewind/replay facility.
#
#    The queue object also holds some (redundant) state information,  which is
#    not currently used (or even accessible),  but can be used to monitor the
#    state of the queue (length,  percentage run,  etc).


# A structure containing information about a movement taking place.
class MoveData:
    def __init__(self, slice=0, dir=0, axis=0):
        self.slice = slice  # int
        self.dir = dir      # 0 or 1.
        self.axis = axis    # int
    def copy(self):
        return MoveData(self.slice, self.dir, self.axis)
    def __str__(self):
        return 'MoveData(slice=%d dir=%d axis=%d)' % (self.slice, self.dir, self.axis)


# A simple compound accessor.

class MoveQueueProgress:
    def __init__(self, current=0, total=0):
        self._current = current
        self._total = total
    @property
    def current(self):
        return self._current
    @property
    def total(self):
        return self._total


# The queue is implemented as a singly-linked list of MoveQueueItem
# objects. For maximum efficiency pushes are made at the back (tail),  pops at
# the front (head),  and the normal direction of recursion is towards the tail;
# retarding the cursor is the one (relatively) expensive operation.
#
# Note that the cursor,  current,  points to the next move to be made (the true
# location of the cursor is between this move and the previous move).

class MoveQueueItem:
    def __init__(self, data=None, mark=0, next=None):
        self.data = data or MoveData() #MoveData
        self.mark = mark    # int
        self.next = next    # MoveQueueItem or None
    def copy(self):
        return MoveQueueItem(self.data.copy(), self.mark, self.next)


class MoveQueue:
    def __init__(self):
        self.head = None        # MoveQueueItem  Pop off this end.
        self.tail = None        # MoveQueueItem  Push on this end.
        self._current = None     # MoveQueueItem  User's 'roaming' cursor.
        self.queue_length = 0   # unsigned      The number of items on the queue.
        self.current_place = 0  # unsigned      Number of steps of current from head.
        self.mark_next_push = 0 # int           Will the next move be a marked point?
    def __del__(self):
        self._release_items(self.head)
    def __str__(self):
        return 'MoveQueue(len=%s)' % self.queue_length

    # Create and return a pointer to a new object,  which represents a queue holding
    # no data.
    def _initialize_new_object (self):
        self.head = self.tail = self._current = None
        self.queue_length = 0
        self.current_place = 0
        self.mark_next_push = 0

    # Release all the memory associated with the given queue object.
    def _release_items(self, next_item):
        while next_item != None:
            next = next_item.next
            next_item.next = None
            next_item = next

    # Routine to push a new MoveData object onto the _back_ of the given queue. A
    # new,  locally-managed copy is made of the incumbent datum. Return 1 (one) if
    # the operation is successful; 0 (zero) indicates that there was insufficient
    # memory to grow the queue.
    def push (self, move_data):
        new_element = MoveQueueItem()
        if new_element == None:
            return 0

        new_element.data = move_data.copy()
        new_element.mark = self.mark_next_push
        new_element.next = None

        if self.tail != None:
            self.tail.next = new_element

        self.tail = new_element
        self.queue_length += 1

        if self.head == None:
            self.head = new_element

        if self._current == None:
            self._current = new_element

        self.mark_next_push = 0
        return 1

    # Procedure to copy the incumbent datum to the current place in the queue,  and
    # drop all subsequent queue entries (so that the current element becomes the
    # tail). If current points past the end of the tail (including the case when
    # there are no data in the queue),  then a regular push operation is performed
    # at the tail,  and in this case zero may be returned if there is insufficient
    # memory to grow the queue. In all other cases 1 (one) will be returned.
    def push_current (self, move_data):
        # If there are no data in the queue,  then perform a standard push
        # operation. Also do this if we are at the tail.
        if self._current == None:
            return self.push(move_data)

        self._release_items(self._current.next)
        self._current.data = move_data.copy()
        self._current.next = None
        self.tail = self._current
        self.queue_length = self.current_place + 1
        self.mark_next_push = 0
        return 1

    # Routine to get the data at the 'current' position in the queue. If there are
    # no data in the queue,  None will be returned.
    def current (self):
        return self._current

    # Routine to retard the current pointer (move it towards the _head_ of the
    # queue). If there are no data on the queue,  or the head already is the current
    # pointer,  then zero is returned to indicate that there are no more data to
    # consider. If a marked item is hit,  -1 is returned.
    #
    # Note that,  as this is implemented as a singly-linked list,  this is the
    # expensive operation,  as we have to trace the list from the head to the
    # current datum,  in order to locate the datum above the current which will
    # become the new current. The alternatives are either to use a doubly-linked
    # list which will make the logic more complicated and use a (tiny) amount of
    # extra memory,  or to use the list in the opposite direction but then pushing,
    # poping and advancing all become expensive operations.
    def retard (self):
        if self.head == self._current:  # We are trying to move past the head.
            return 0

        elif self._current == None:         # We are past the tail.
            self._current = self.tail

        else:        # We must be somewhere in the middle.
            cursor = self.head
            while cursor.next != self._current:
                 cursor = cursor.next
            self._current = cursor

        self.current_place -= 1

        return -1 if self._current.mark == 1 else 1

    # Remove the current object and all those that come afterwards.
    def truncate (self):
        if self._current == self.head:
            self._release_items(self.head)
            self._initialize_new_object ()

        else:
            self.mark_next_push = (
                0 if self._current == None else self._current.mark)

            self.retard()

            if self._current != None:
                self._release_items(self._current.next)

            self._current.next = None
            self.tail = self._current
            self.advance()
            self.queue_length = self.current_place

    # Routine to advance the current position (move it towards the _tail_). If the
    # current position is already at the tail (which will also be the case if the
    # queue is empty) then no action takes place and 0 (zero) is returned to
    # indicate that no more data are available for consideration.
    def advance (self):
        if self._current == None:
            return 0

        self._current = self._current.next
        self.current_place += 1

        return self._current != None

    # A simple compound accessor.
    def progress(self):
        return MoveQueueProgress(self.current_place, self.queue_length)

    # And a mutator.
    def mark_current(self, mark=True):
        if self._current:
            self._current.mark = 1 if mark else 0
        else:
            self.mark_next_push = 1 if mark else 0

